# Standard library imports
import functools
import warnings

# Local imports
from uplink import (
    arguments,
    auth as auth_,
    clients,
    converters as converters_,
    exceptions,
    helpers,
    hooks as hooks_,
    interfaces,
    session,
    utils,
)
from uplink.clients import io

__all__ = ["build", "Consumer"]


class RequestPreparer(object):
    def __init__(self, builder, consumer=None):
        self._client = builder.client
        self._base_url = str(builder.base_url)
        self._converters = list(builder.converters)
        self._auth = builder.auth
        self._consumer = consumer

        if builder.hooks:
            self._session_chain = hooks_.TransactionHookChain(*builder.hooks)
        else:
            self._session_chain = None

    @staticmethod
    def _get_request_hooks(contract):
        chain = list(contract.transaction_hooks)
        if callable(contract.return_type):
            chain.append(hooks_.ResponseHandler(contract.return_type))
        return chain

    def _wrap_hook(self, func):
        return functools.partial(func, self._consumer)

    def apply_hooks(self, execution_builder, chain):
        # TODO:
        #   Instead of creating a TransactionChain, we could simply
        #   add each response and error handler in the chain to the
        #   execution builder. This would allow heterogenous response
        #   and error handlers. Right now, the TransactionChain
        #   enforces that all response/error handlers are blocking if
        #   any response/error handler is blocking, which is
        #   unnecessary now that we delegate execution to an IO layer.
        if chain.handle_response is not None:
            execution_builder.with_callbacks(
                self._wrap_hook(chain.handle_response)
            )
        execution_builder.with_errbacks(self._wrap_hook(chain.handle_exception))

    def prepare_request(self, request_builder, execution_builder):
        self._auth(request_builder)
        request_hooks = self._get_request_hooks(request_builder)
        if request_hooks:
            chain = hooks_.TransactionHookChain(*request_hooks)
            chain.audit_request(self._consumer, request_builder)
            self.apply_hooks(execution_builder, chain)
        if self._session_chain:
            self.apply_hooks(execution_builder, self._session_chain)

        execution_builder.with_client(self._client)
        execution_builder.with_io(self._client.io())
        execution_builder.with_template(request_builder.request_template)

    def create_request_builder(self, definition):
        registry = definition.make_converter_registry(self._converters)
        req = helpers.RequestBuilder(self._client, registry, self._base_url)
        if self._session_chain:
            self._session_chain.audit_request(self._consumer, req)
        return req


class CallFactory(object):
    def __init__(
        self, request_preparer, request_definition, execution_builder_factory
    ):
        self._request_preparer = request_preparer
        self._request_definition = request_definition
        self._execution_builder_factory = execution_builder_factory

    def __call__(self, *args, **kwargs):
        request_builder = self._request_preparer.create_request_builder(
            self._request_definition
        )
        self._request_definition.define_request(request_builder, args, kwargs)
        execution_builder = self._execution_builder_factory()
        self._request_preparer.prepare_request(
            request_builder, execution_builder
        )
        execution = execution_builder.build()
        return execution.start(
            # TODO: Create request value object
            (request_builder.method, request_builder.url, request_builder.info)
        )


class Builder(interfaces.CallBuilder):
    """The default callable builder."""

    def __init__(self):
        self._base_url = ""
        self._hooks = []
        self._client = clients.get_client()
        self._converters = converters_.get_default_converter_factories()
        self._auth = auth_.get_auth()

    @property
    def client(self):
        return self._client

    @client.setter
    def client(self, client):
        if client is not None:
            self._client = clients.get_client(client)

    @property
    def hooks(self):
        return iter(self._hooks)

    def add_hook(self, *hooks):
        self._hooks.extend(hooks)

    @property
    def base_url(self):
        return self._base_url

    @base_url.setter
    def base_url(self, base_url):
        self._base_url = base_url

    @property
    def converters(self):
        return self._converters

    @converters.setter
    def converters(self, converters):
        if isinstance(converters, converters_.interfaces.Factory):
            converters = (converters,)
        self._converters = tuple(converters)
        self._converters += converters_.get_default_converter_factories()

    @property
    def auth(self):
        return self._auth

    @auth.setter
    def auth(self, auth):
        if auth is not None:
            self._auth = auth_.get_auth(auth)

    def build(self, definition, consumer=None):
        """
        Creates a callable that uses the provided definition to execute
        HTTP requests when invoked.
        """
        return CallFactory(
            RequestPreparer(self, consumer),
            definition,
            io.RequestExecutionBuilder,
        )


class ConsumerMethod(object):
    """
    A wrapper around a :py:class`interfaces.RequestDefinitionBuilder`
    instance bound to a :py:class:`Consumer` subclass, mainly responsible
    for controlling access to the instance.
    """

    def __init__(self, owner_name, attr_name, request_definition_builder):
        self._request_definition_builder = request_definition_builder
        self._owner_name = owner_name
        self._attr_name = attr_name
        self._request_definition = self._build_definition()

    def _build_definition(self):
        try:
            return self._request_definition_builder.build()
        except exceptions.InvalidRequestDefinition as error:
            # TODO: Find a Python 2.7 compatible way to reraise
            raise exceptions.UplinkBuilderError(
                self._owner_name, self._attr_name, error
            )

    def __get__(self, instance, owner):
        # TODO:
        #   Consider caching by instance/owner using WeakKeyDictionary.
        #   This will avoid the extra copy/create per attribute reference.
        #   However, we should do this after investigating for any latent cases
        #   of unnecessary overhead in the codebase as a whole.
        if instance is None:
            # This code path is traditionally called when applying a class
            # decorator to a Consumer. We should return a copy of the definition
            # builder to avoid class decorators on a subclass from polluting
            # other siblings (#152).
            value = self._request_definition_builder.copy()
        else:
            value = instance.session.create(instance, self._request_definition)

        # Make the return value look like the original method (e.g., inherit
        # docstrings and other function attributes).
        # TODO: Ideally, we should wrap once instead of on each reference.
        self._request_definition_builder.update_wrapper(value)
        return value


class ConsumerMeta(type):
    @staticmethod
    def _wrap_if_definition(cls_name, key, value):
        wrapped_value = value
        if isinstance(value, interfaces.RequestDefinitionBuilder):
            wrapped_value = ConsumerMethod(cls_name, key, value)
            value.update_wrapper(wrapped_value)
        return wrapped_value

    @staticmethod
    def _set_init_handler(namespace):
        try:
            init = namespace["__init__"]
        except KeyError:
            pass
        else:
            builder = arguments.ArgumentAnnotationHandlerBuilder.from_func(init)
            handler = builder.build()

            @functools.wraps(init)
            def new_init(self, *args, **kwargs):
                init(self, *args, **kwargs)
                call_args = utils.get_call_args(init, self, *args, **kwargs)
                f = functools.partial(
                    handler.handle_call_args, call_args=call_args
                )
                hook = hooks_.RequestAuditor(f)
                self.session.inject(hook)

            namespace["__init__"] = new_init

    def __new__(mcs, name, bases, namespace):
        mcs._set_init_handler(namespace)

        # Wrap all definition builders with a special descriptor that
        # handles attribute access behavior.
        for key, value in namespace.items():
            namespace[key] = mcs._wrap_if_definition(name, key, value)
        return super(ConsumerMeta, mcs).__new__(mcs, name, bases, namespace)

    def __setattr__(cls, key, value):
        value = cls._wrap_if_definition(cls.__name__, key, value)
        super(ConsumerMeta, cls).__setattr__(key, value)


_Consumer = ConsumerMeta("_Consumer", (), {})


class Consumer(interfaces.Consumer, _Consumer):
    """
    Base consumer class with which to define custom consumers.

    Example usage:

    .. code-block:: python

        from uplink import Consumer, get

        class GitHub(Consumer):

            @get("/users/{user}")
            def get_user(self, user):
                pass

        client = GitHub("https://api.github.com/")
        client.get_user("prkumar").json()  # {'login': 'prkumar', ... }

    Args:
        base_url (:obj:`str`, optional): The base URL for any request
            sent from this consumer instance.
        client (optional): A supported HTTP client instance (e.g.,
            a :class:`requests.Session`) or an adapter (e.g.,
            :class:`~uplink.RequestsClient`).
        converters (:class:`ConverterFactory`, optional):
            One or more objects that encapsulate custom
            (de)serialization strategies for request properties and/or
            the response body. (E.g.,
            :class:`~uplink.converters.MarshmallowConverter`)
        auth (:obj:`tuple` or :obj:`callable`, optional): The
            authentication object for this consumer instance.
        hooks (:class:`~uplink.hooks.TransactionHook`, optional):
            One or more hooks to modify behavior of request execution
            and response handling (see :class:`~uplink.response_handler`
            or :class:`~uplink.error_handler`).
    """

    def __init__(
        self,
        base_url="",
        client=None,
        converters=(),
        auth=None,
        hooks=(),
        **kwargs
    ):
        builder = Builder()
        builder.base_url = base_url
        builder.converters = kwargs.pop("converter", converters)
        hooks = kwargs.pop("hook", hooks)
        if isinstance(hooks, hooks_.TransactionHook):
            hooks = (hooks,)
        builder.add_hook(*hooks)
        builder.auth = auth
        builder.client = client
        self.__session = session.Session(builder)
        self.__client = builder.client

    def _inject(self, hook, *more_hooks):
        self.session.inject(hook, *more_hooks)

    @property
    def session(self):
        """
        The :class:`~uplink.session.Session` object for this consumer
        instance.

        Exposes the configuration of this :class:`~uplink.Consumer`
        instance and allows the persistence of certain properties across
        all requests sent from that instance.

        Example usage:

        .. code-block:: python

            import uplink

            class MyConsumer(uplink.Consumer):
                def __init__(self, language):
                    # Set this header for all requests of the instance.
                    self.session.headers["Accept-Language"] = language
                    ...

        Returns:
            :class:`~uplink.session.Session`
        """
        return self.__session

    @property
    def exceptions(self):
        """
        An enum of standard HTTP client exceptions that can be handled.

        This property enables the handling of specific exceptions from
        the backing HTTP client.

        Example:

            .. code-block:: python

                try:
                    github.get_user(user_id)
                except github.exceptions.ServerTimeout:
                    # Handle the timeout of the request
                    ...
        """
        return self.__client.exceptions


def build(service_cls, *args, **kwargs):
    name = service_cls.__name__
    warnings.warn(
        "`uplink.build` is deprecated and will be removed in v1.0.0. "
        "To construct a consumer instance, have `{0}` inherit "
        "`uplink.Consumer` then instantiate (e.g., `{0}(...)`). ".format(name),
        DeprecationWarning,
    )
    consumer = type(name, (service_cls, Consumer), dict(service_cls.__dict__))
    return consumer(*args, **kwargs)



__all__ = ["sleep", "send", "finish", "fail", "prepare"]


def sleep(seconds):
    """
    Transitions the execution to pause for the allotted duration.

    Args:
        seconds: The number of seconds to delay execution.
    """

    def action(state):
        return state.sleep(seconds)

    return action


def send(request):
    """
    Transitions the execution to send the given request.

    Args:
        request: The intended request data to be sent.
    """

    def action(state):
        return state.send(request)

    return action


def finish(response):
    """
    Transitions the execution to completion.

    Args:
        response: The object to return to the execution's invoker.
    """

    def action(state):
        return state.finish(response)

    return action


def fail(exc_type, exc_val, exc_tb):
    """
    Transitions the execution to fail with a specific error.

    This will prompt the execution of any RequestTemplate.after_exception hooks.

    Args:
        exc_type: The exception class.
        exc_val: The exception object.
        exc_tb: The exception's stacktrace.
    """

    def action(state):
        return state.fail(exc_type, exc_val, exc_tb)

    return action


def prepare(request):
    """
    Transitions the execution to prepare the given request.

    This will prompt the execution of any RequestTemplate.before_request.

    Args:
        request: The intended request data to be sent.
    """

    def action(state):
        return state.prepare(request)

    return action



# Standard imports
import random
import sys

# Constants
MAX_VALUE = sys.maxsize / 2

__all__ = ["jittered", "exponential", "fixed"]


def jittered(base=2, multiplier=1, minimum=0, maximum=MAX_VALUE):
    """
    Waits using capped exponential backoff and full jitter.

    The implementation is discussed in `this AWS Architecture Blog
    post <https://amzn.to/2xc2nK2>`_, which recommends this approach
    for any remote clients, as it minimizes the total completion
    time of competing clients in a distributed system experiencing
    high contention.
    """
    exp_backoff = exponential(base, multiplier, minimum, maximum)
    return lambda *_: iter(
        random.uniform(0, 1) * delay for delay in exp_backoff()
    )  # pragma: no cover


def exponential(base=2, multiplier=1, minimum=0, maximum=MAX_VALUE):
    """
    Waits using capped exponential backoff, meaning that the delay
    is multiplied by a constant ``base`` after each attempt, up to
    an optional ``maximum`` value.
    """

    def wait_iterator():
        delay = multiplier
        while minimum > delay:
            delay *= base
        while True:
            yield min(delay, maximum)
            delay *= base

    return wait_iterator


def fixed(seconds):
    """Waits for a fixed number of ``seconds`` before each retry."""

    def wait_iterator():
        while True:
            yield seconds

    return wait_iterator



# Standard library imports
import collections

# Third-party imports
import pytest

# Local imports
from uplink import decorators, interfaces


@pytest.fixture
def method_annotation():
    return decorators.MethodAnnotation()


@pytest.fixture
def method_annotation_mock(mocker):
    return mocker.Mock(spec=decorators.MethodAnnotation)


@pytest.fixture
def method_handler_builder():
    return decorators.MethodAnnotationHandlerBuilder()


class TestMethodAnnotationHandlerBuilder(object):
    def test_add_annotation(
        self, mocker, method_handler_builder, method_annotation_mock
    ):
        method_handler_builder.listener = mocker.stub()
        method_handler_builder.add_annotation(method_annotation_mock)
        method_handler_builder.listener.assert_called_with(
            method_annotation_mock
        )
        annotations = method_handler_builder.build().annotations
        assert list(annotations) == [method_annotation_mock]

    def test_build(self, method_handler_builder, method_annotation):
        method_handler_builder.add_annotation(method_annotation)
        handler = method_handler_builder.build()
        assert list(handler.annotations) == [method_annotation]

    def test_class_level_appears_before_method_level_annotations(
        self, method_handler_builder
    ):
        method_level1 = decorators.MethodAnnotation()
        method_level2 = decorators.MethodAnnotation()
        class_level1 = decorators.MethodAnnotation()
        class_level2 = decorators.MethodAnnotation()
        method_handler_builder.add_annotation(method_level1)
        method_handler_builder.add_annotation(method_level2)
        method_handler_builder.add_annotation(class_level1, is_class=True)
        method_handler_builder.add_annotation(class_level2, is_class=True)
        handler = method_handler_builder.build()
        assert list(handler.annotations) == [
            class_level1,
            class_level2,
            method_level1,
            method_level2,
        ]


class TestMethodAnnotationHandler(object):
    def test_handle_builder(self, request_builder, method_annotation_mock):
        handler = decorators.MethodAnnotationHandler([method_annotation_mock])
        handler.handle_builder(request_builder)
        method_annotation_mock.modify_request.assert_called_with(
            request_builder
        )

    def test_annotations(self, method_annotation_mock):
        handler = decorators.MethodAnnotationHandler([method_annotation_mock])
        assert list(handler.annotations) == [method_annotation_mock]


class TestMethodAnnotation(object):
    class FakeMethodAnnotation(decorators.MethodAnnotation):
        _can_be_static = True

    def test_call_with_class(
        self, method_annotation, request_definition_builder
    ):
        class Class(interfaces.Consumer):
            builder = request_definition_builder

        method_annotation(Class)
        builder = request_definition_builder.method_handler_builder
        builder.add_annotation.assert_called_with(
            method_annotation, is_class=True
        )

    def test_static_call_with_class(self, mocker, request_definition_builder):
        class Class(interfaces.Consumer):
            builder = request_definition_builder

        self.FakeMethodAnnotation(Class)
        builder = request_definition_builder.method_handler_builder
        builder.add_annotation.assert_called_with(mocker.ANY, is_class=True)

    def test_call_with_builder(
        self, method_annotation, request_definition_builder
    ):
        method_annotation(request_definition_builder)
        builder = request_definition_builder.method_handler_builder
        builder.add_annotation.assert_called_with(method_annotation)

    def test_static_call_with_builder(self, mocker, request_definition_builder):
        self.FakeMethodAnnotation(request_definition_builder)
        builder = request_definition_builder.method_handler_builder
        builder.add_annotation.assert_called_with(mocker.ANY)

    def test_method_in_http_method_blacklist(self, request_definition_builder):
        class DummyAnnotation(decorators.MethodAnnotation):
            _http_method_blacklist = ["GET"]

        request_definition_builder.method = "GET"
        assert not DummyAnnotation.supports_http_method(
            request_definition_builder.method
        )

    def test_method_not_in_http_method_blacklist(
        self, request_definition_builder
    ):
        class DummyAnnotation(decorators.MethodAnnotation):
            _http_method_whitelist = ["POST"]

        request_definition_builder.method = "POST"
        request_definition_builder.__name__ = "dummy"
        assert DummyAnnotation().supports_http_method(
            request_definition_builder.method
        )

    def test_call_with_child_class(
        self, method_annotation, request_definition_builder
    ):
        class Parent(interfaces.Consumer):
            builder = request_definition_builder

        class Child(Parent):
            pass

        # Method annotation should decorate RequestDefinitionBuilder attribute
        # of parent class (e.g., `Parent.builder`).
        method_annotation(Child)
        builder = request_definition_builder.method_handler_builder
        assert builder.add_annotation.called

    def test_no_call_on_non_consumer(
        self, method_annotation, request_definition_builder
    ):
        class Class(object):
            builder = request_definition_builder

        method_annotation(Class)
        builder = request_definition_builder.method_handler_builder
        assert not builder.add_annotation.called


# TODO: Refactor test cases for method annotations into test case class.


def test_headers(request_builder):
    headers = decorators.headers({"key_1": "value_1"})
    headers.modify_request(request_builder)
    assert request_builder.info["headers"] == {"key_1": "value_1"}

    headers_str = decorators.headers("key_1: value_1")
    headers_str.modify_request(request_builder)
    assert request_builder.info["headers"] == {"key_1": "value_1"}

    headers_lst = decorators.headers(["key_1: value_1", "key_2: value_2"])
    headers_lst.modify_request(request_builder)
    header = {"key_1": "value_1", "key_2": "value_2"}
    assert request_builder.info["headers"] == header


def test_params(request_builder):
    params = decorators.params({"param": "value"})
    params.modify_request(request_builder)
    assert request_builder.info["params"] == {"param": "value"}


def test_params_with_string(request_builder):
    params_with_str = decorators.params("param1=value1&param2=value2")
    params_with_str.modify_request(request_builder)
    assert request_builder.info["params"] == {
        "param1": "value1",
        "param2": "value2",
    }


def test_params_with_list(request_builder):
    params_list = decorators.params(["param1=value1", "param2=value2"])
    params_list.modify_request(request_builder)
    assert request_builder.info["params"] == {
        "param1": "value1",
        "param2": "value2",
    }


def test_form_url_encoded(request_builder):
    form_url_encoded = decorators.form_url_encoded()
    form_url_encoded.modify_request(request_builder)
    assert "headers" not in request_builder.info


def test_multipart(request_builder):
    multipart = decorators.multipart()
    multipart.modify_request(request_builder)
    assert "headers" not in request_builder.info


def test_json(request_builder):
    json = decorators.json()

    # Verify that we add the hook
    json.modify_request(request_builder)
    request_builder.add_transaction_hook.assert_called_with(json._hook)

    # Verify
    request_builder.info["data"] = {"field_name": "field_value"}
    json.set_json_body(request_builder)
    assert request_builder.info["json"] == {"field_name": "field_value"}
    assert len(request_builder.info["json"]) == 1

    # Check delete
    del request_builder.info["json"]["field_name"]
    assert len(request_builder.info["json"]) == 0
    assert "data" not in request_builder.info

    # Verify nested attribute
    request_builder.info["data"] = {("outer", "inner"): "inner_value"}
    json.set_json_body(request_builder)
    assert request_builder.info["json"] == {"outer": {"inner": "inner_value"}}
    assert "data" not in request_builder.info

    # Verify that error is raised when path is empty
    request_builder.info["data"] = {(): "value"}
    with pytest.raises(ValueError):
        json.set_json_body(request_builder)

    # Verify that error is raised when paths conflict
    request_builder.info["data"] = body = collections.OrderedDict()
    body["key"] = "outer"
    body["key", "inner"] = "inner value"
    with pytest.raises(ValueError):
        json.set_json_body(request_builder)


def test_json_list(request_builder):
    json = decorators.json()

    # Verify nested attribute
    request_builder.info["data"] = ["one", "two", "three"]
    json.set_json_body(request_builder)
    assert request_builder.info["json"] == ["one", "two", "three"]
    assert "data" not in request_builder.info


def test_timeout(request_builder):
    timeout = decorators.timeout(60)
    timeout.modify_request(request_builder)
    assert request_builder.info["timeout"] == 60


def test_args(request_definition_builder):
    args = decorators.args(str, str, name=str)
    args.modify_request_definition(request_definition_builder)
    builder = request_definition_builder.argument_handler_builder
    builder.set_annotations.assert_called_with((str, str), name=str)


def test_args_decorate_function(mocker):
    handler = mocker.Mock(spec=["set_annotations"])

    @classmethod
    def patched(*_):
        return handler

    mocker.patch(
        "uplink.arguments.ArgumentAnnotationHandlerBuilder.from_func", patched
    )
    args = decorators.args(str, str, name=str)
    func = mocker.stub()
    args(func)
    handler.set_annotations.assert_called_with((str, str), name=str)


def test_args_call_old(request_definition_builder):
    annotation = decorators.args(str, str, name=str)
    annotation(request_definition_builder)
    handler = request_definition_builder.method_handler_builder
    handler.add_annotation.assert_called_with(annotation)


def test_response_handler(request_builder):
    @decorators.response_handler
    def handler(response):
        return response

    handler.modify_request(request_builder)
    request_builder.add_transaction_hook.assert_called_with(handler)


def test_error_handler(request_builder):
    @decorators.error_handler
    def handler(*_):
        return True

    handler.modify_request(request_builder)
    request_builder.add_transaction_hook.assert_called_with(handler)


def test_inject(request_builder, transaction_hook_mock):
    handler = decorators.inject(transaction_hook_mock)
    handler.modify_request(request_builder)
    request_builder.add_transaction_hook.assert_called_with(handler)



# Third-party imports
import asyncio
import sys

# Local models
from uplink.clients.io import interfaces

__all__ = ["AsyncioStrategy"]


class AsyncioStrategy(interfaces.IOStrategy):
    """A non-blocking execution strategy using asyncio."""

    @asyncio.coroutine
    def invoke(self, func, args, kwargs, callback):
        try:
            response = yield from func(*args, **kwargs)
        except Exception as error:
            tb = sys.exc_info()[2]
            response = yield from callback.on_failure(type(error), error, tb)
        else:
            response = yield from callback.on_success(response)
        return response

    @asyncio.coroutine
    def sleep(self, duration, callback):
        yield from asyncio.sleep(duration)
        response = yield from callback.on_success()
        return response

    @asyncio.coroutine
    def finish(self, response):
        yield
        return response

    @asyncio.coroutine
    def execute(self, executable):
        response = yield from executable.execute()
        return response



# Local imports
from uplink import returns


def test_returns(request_builder):
    custom = returns(str)
    request_builder.get_converter.return_value = str
    request_builder.return_type = returns.ReturnType.with_decorator(
        None, custom
    )
    custom.modify_request(request_builder)
    assert request_builder.return_type(2) == "2"


def test_returns_with_multiple_decorators(request_builder, mocker):
    decorator1 = returns(str)
    decorator2 = returns.json()
    request_builder.get_converter.return_value = str
    first_type = returns.ReturnType.with_decorator(None, decorator1)
    second_type = (
        request_builder.return_type
    ) = returns.ReturnType.with_decorator(
        first_type, decorator2
    )

    # Verify that the return type doesn't change after being handled by first decorator
    decorator1.modify_request(request_builder)
    assert request_builder.return_type is second_type

    # Verify that the second decorator does handle the return type
    mock_response = mocker.Mock()
    mock_response.json.return_value = {"key": "value"}
    decorator2.modify_request(request_builder)
    assert request_builder.return_type(mock_response) == str(
        mock_response.json()
    )


def test_returns_json(request_builder, mocker):
    mock_response = mocker.Mock()
    mock_response.json.return_value = {"key": "value"}
    request_builder.get_converter.return_value = str
    returns_json = returns.json(str, ())
    request_builder.return_type = returns.ReturnType.with_decorator(
        None, returns_json
    )
    returns_json.modify_request(request_builder)
    assert isinstance(request_builder.return_type, returns.ReturnType)
    assert callable(request_builder.return_type)
    assert request_builder.return_type(mock_response) == str(
        mock_response.json()
    )

    # Verify: Idempotent
    returns_json.modify_request(request_builder)
    assert isinstance(request_builder.return_type, returns.ReturnType)
    assert callable(request_builder.return_type)
    assert request_builder.return_type(mock_response) == str(
        mock_response.json()
    )

    # Verify: Doesn't apply to unsupported types
    request_builder.get_converter.return_value = None
    returns_json = returns.json(str, ())
    request_builder.return_type = returns.ReturnType.with_decorator(
        None, returns_json
    )
    returns_json.modify_request(request_builder)
    assert not callable(request_builder.return_type)


def test_returns_JsonStrategy(mocker):
    response = mocker.Mock(spec=["json"])
    response.json.return_value = {"hello": "world"}
    converter = returns.JsonStrategy(lambda x: x, "hello")
    assert converter(response) == "world"

    converter = returns.JsonStrategy(lambda y: y + "!", "hello")
    assert converter(response) == "world!"

    assert returns.JsonStrategy(1).unwrap() == 1
